/*
 *  Copyright (c) 2018, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 *  WSO2 Inc. licenses this file to you under the Apache License,
 *  Version 2.0 (the "License"); you may not use this file except
 *  in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.ballerinalang.net.http.serviceendpoint;

import org.ballerinalang.bre.Context;
import org.ballerinalang.connector.api.BLangConnectorSPIUtil;
import org.ballerinalang.connector.api.ParamDetail;
import org.ballerinalang.connector.api.Service;
import org.ballerinalang.connector.api.Struct;
import org.ballerinalang.model.types.TypeKind;
import org.ballerinalang.natives.annotations.Argument;
import org.ballerinalang.natives.annotations.BallerinaFunction;
import org.ballerinalang.natives.annotations.Receiver;
import org.ballerinalang.net.http.BallerinaHTTPConnectorListener;
import org.ballerinalang.net.http.HTTPServicesRegistry;
import org.ballerinalang.net.http.HttpConnectorPortBindingListener;
import org.ballerinalang.net.http.HttpConstants;
import org.ballerinalang.net.http.WebSocketConstants;
import org.ballerinalang.net.http.WebSocketServerConnectorListener;
import org.ballerinalang.net.http.WebSocketService;
import org.ballerinalang.net.http.WebSocketServicesRegistry;
import org.ballerinalang.util.exceptions.BallerinaException;
import org.wso2.transport.http.netty.contract.ServerConnector;
import org.wso2.transport.http.netty.contract.ServerConnectorFuture;

import static org.ballerinalang.net.http.HttpConstants.HTTP_LISTENER_ENDPOINT;

/**
 * Get the ID of the connection.
 *
 * @since 0.966
 */

@BallerinaFunction(
        orgName = "ballerina", packageName = "http",
        functionName = "register",
        receiver = @Receiver(type = TypeKind.OBJECT, structType = HTTP_LISTENER_ENDPOINT,
                structPackage = "ballerina/http"),
        args = {@Argument(name = "serviceType", type = TypeKind.SERVICE),
                @Argument(name = "annotationData", type = TypeKind.MAP)},
        isPublic = true
)
public class Register extends AbstractHttpNativeFunction {

    @Override
    public void execute(Context context) {
        Service service = BLangConnectorSPIUtil.getServiceRegistered(context);
        Struct serviceEndpoint = BLangConnectorSPIUtil.getConnectorEndpointStruct(context);

        HTTPServicesRegistry httpServicesRegistry = getHttpServicesRegistry(serviceEndpoint);
        WebSocketServicesRegistry webSocketServicesRegistry = getWebSocketServicesRegistry(serviceEndpoint);

        ParamDetail param;
        if (service.getResources().length > 0 && (param = service.getResources()[0].getParamDetails().get(0)) != null) {
            String callerType = param.getVarType().toString();
            if (HttpConstants.HTTP_CALLER_NAME.equals(callerType)) {
                httpServicesRegistry.registerService(service);
            } else if (WebSocketConstants.WEBSOCKET_CALLER_NAME.equals(callerType)) {
                WebSocketService webSocketService = new WebSocketService(service);
                webSocketServicesRegistry.registerService(webSocketService);
            }
        } else {
            httpServicesRegistry.registerService(service);
        }
        // TODO: 11/24/18 failure need to be validated in the compile time

        if (!isConnectorStarted(serviceEndpoint)) {
            startServerConnector(serviceEndpoint, httpServicesRegistry, webSocketServicesRegistry);
        }
        context.setReturnValues();
    }

    private void startServerConnector(Struct serviceEndpoint, HTTPServicesRegistry httpServicesRegistry,
                                      WebSocketServicesRegistry webSocketServicesRegistry) {
        ServerConnector serverConnector = getServerConnector(serviceEndpoint);
        ServerConnectorFuture serverConnectorFuture = serverConnector.start();
        serverConnectorFuture.setHttpConnectorListener(
                new BallerinaHTTPConnectorListener(httpServicesRegistry, serviceEndpoint
                        .getStructField(HttpConstants.SERVICE_ENDPOINT_CONFIG)));
        serverConnectorFuture.setWebSocketConnectorListener(new WebSocketServerConnectorListener(
                webSocketServicesRegistry, serviceEndpoint.getStructField(HttpConstants.SERVICE_ENDPOINT_CONFIG)));

        serverConnectorFuture.setPortBindingEventListener(new HttpConnectorPortBindingListener());
        try {
            serverConnectorFuture.sync();
        } catch (Exception ex) {
            throw new BallerinaException("failed to start server connector '" + serverConnector.getConnectorID()
                    + "': " + ex.getMessage(), ex);
        }
        serviceEndpoint.addNativeData(HttpConstants.CONNECTOR_STARTED, true);
    }
}
